package excel

import (
	"encoding/json"
	"errors"
	"eta_gn/eta_chart_lib/models"
	"eta_gn/eta_chart_lib/models/data_manage"
	"eta_gn/eta_chart_lib/models/request"
	"eta_gn/eta_chart_lib/services/data"
	"eta_gn/eta_chart_lib/utils"
	"fmt"
	"sort"
	"strconv"
	"strings"
	"time"

	"github.com/shopspring/decimal"
	"github.com/yidane/formula"
)

type BaseCalculate struct {
	DataList      []*models.EdbDataList
	Frequency     string `description:"需要转换的频度"`
	Formula       interface{}
	Calendar      string `description:"公历/农历"`
	MoveType      int    `description:"移动方式:1:领先(默认),2:滞后"`
	MoveFrequency string `description:"移动频度"`
	FromFrequency string `description:"来源的频度"`
	Source        int    `description:"1：累计值转月；2：累计值转季；3：同比值；4：同差值；5：N数值移动平均数计算；6：环比值；7：环差值；8：升频；9：降频；10：时间移位；11：超季节性；12：年化；13：累计值；14：累计值年初至今；15：指数修匀；16：日均值"`
}

type Cell struct {
	Column   int                           `description:"行"`
	Row      int                           `description:"列"`
	CellInfo request.MixedTableCellDataReq `description:"对应的单元格信息"`
}

func GetMixedTableCellData(mixedTableReq request.MixedTableReq, lang string) (newMixedTableCellDataList [][]request.MixedTableCellDataReq, err error, errMsg string) {
	cellRelationConf := mixedTableReq.CellRelation
	config := mixedTableReq.Data

	cellRelationConfMap := make(map[string]request.CellRelationConf)
	cellRelationConfList := make([]request.CellRelationConf, 0)
	if cellRelationConf != `` {
		err = json.Unmarshal([]byte(cellRelationConf), &cellRelationConfList)
		if err != nil {
			return
		}

		for _, v := range cellRelationConfList {
			cellRelationConfMap[v.Key] = v
		}
	}

	config, edbInfoIdList, _, err, errMsg := handleConfig(config)
	if err != nil {
		return
	}

	edbInfoList, err := data_manage.GetEdbInfoByIdList(edbInfoIdList)
	if err != nil {
		return
	}

	edbInfoMap := make(map[int]*data_manage.EdbInfo)
	edbDayDataListMap := make(map[int]map[string]float64)
	edbMonthDataListMap := make(map[int]map[string]string)
	edbDataListMap := make(map[int][]*models.EdbDataList)
	for _, edbInfo := range edbInfoList {
		edbInfoMap[edbInfo.EdbInfoId] = edbInfo

		dataList := make([]*models.EdbDataList, 0)
		switch edbInfo.EdbInfoType {
		case 0:
			dataList, _ = models.GetEdbDataList(edbInfo.Source, edbInfo.SubSource, edbInfo.EdbInfoId, ``, ``)
		case 1:
			_, dataList, _, _, _, _ = data.GetPredictDataListByPredictEdbInfoId(edbInfo.EdbInfoId, ``, ``, false)
		default:
			err = errors.New(fmt.Sprint("获取失败,指标类型异常", edbInfo.EdbInfoType))
		}

		dateValMap := make(map[string]float64)
		monthDateMap := make(map[string]string)
		for _, tmpData := range dataList {
			dateValMap[tmpData.DataTime] = tmpData.Value
			yearMonth := strings.Join(strings.Split(tmpData.DataTime, "-")[0:2], "-")
			if _, ok := monthDateMap[yearMonth]; !ok {
				monthDateMap[yearMonth] = tmpData.DataTime
			}
		}
		edbDayDataListMap[edbInfo.EdbInfoId] = dateValMap
		edbMonthDataListMap[edbInfo.EdbInfoId] = monthDateMap
		edbDataListMap[edbInfo.EdbInfoId] = dataList
	}

	cellDataRelationMap := make(map[string]request.MixedTableCellDataReq, 0)

	for k, row := range config {
		for i, cell := range row {
			if cell.DataType == request.DateDT && cell.DataTimeType == request.EdbDateDT {
				var edbDateConfig request.EdbDateConf
				err = json.Unmarshal([]byte(cell.Value), &edbDateConfig)
				if err != nil {
					return
				}
				if dataList, ok := edbDataListMap[edbDateConfig.EdbInfoId]; ok {
					var newDate string
					newDate, err = GetEdbDateByMoveForward(config[k][i].Value, dataList)
					if err != nil {
						return
					}
					newDate, err = HandleMixTableDateChange(newDate, config[k][i].Value)
					if err != nil {
						return
					}
					cell.ShowValue = newDate
					cell.DataTime = newDate
					config[k][i] = cell
				}
			}
			row[i] = cell

			cellDataRelationMap[cell.Uid] = cell
		}
		config[k] = row
	}

	edbSourceDataMap := make(map[string]data.BaseCalculateDataResp)

	cellKeyVal := make(map[string]float64)

	calculateCellMap := make(map[string]Cell)
	calculateChainList := make([]string, 0)
	dateCalculateList := make([]string, 0)
	showStyleList := make([]string, 0)

	for k, row := range config {
		for i, cell := range row {
			switch cell.DataType {
			case request.EdbDT: // 指标类型
				if cell.Value == `` {
					if edbInfo, ok := edbInfoMap[cell.EdbInfoId]; ok {
						cell.ShowValue = edbInfo.EdbName
					}
				} else {
					cell.ShowValue = cell.Value
				}
			case request.InsertDataDT, request.PopInsertDataDT: // 数据类型
				cell.ShowValue = ``

				if relationConf, ok := cellRelationConfMap[cell.Uid]; ok { //表示表格日期
					if relationConf.RelationDate.Key == `` {
						continue
					}
					relationCell, ok := cellDataRelationMap[relationConf.RelationDate.Key]
					if !ok {
						continue
					}

					tmpDateValMap := make(map[string]float64)
					if dateValMap, ok := edbDayDataListMap[cell.EdbInfoId]; ok {
						tmpDateValMap = dateValMap
					}
					relationDate := relationCell.DataTime
					if strings.Contains(cell.Value, "{") {
						relationDate, err = HandleMixTableDateChange(relationDate, cell.Value)
						if err != nil {
							return
						}
					} else {
						cell.Value = ""
					}

					if val, ok2 := tmpDateValMap[relationDate]; ok2 {
						cellKeyVal[cell.Uid] = val
						cell.ShowValue = utils.FormatMixTableDataShowValue(val)
					}
				} else {
					if cell.DataTime == `` {
						if dateValList, ok := edbDataListMap[cell.EdbInfoId]; ok {
							tmpLenData := len(dateValList)
							if tmpLenData > 0 {
								if !strings.Contains(cell.Value, "{") {
									cell.Value = ""
								}
								var newDate string
								newDate, err = GetEdbDateByMoveForward(cell.Value, dateValList)
								if err != nil {
									return
								}
								newDate, err = HandleMixTableDateChange(newDate, cell.Value)
								if err != nil {
									return
								}
								var finalVal string
								for _, v := range dateValList {
									if v.DataTime == newDate {
										finalVal = utils.FormatMixTableDataShowValue(v.Value)
										cellKeyVal[cell.Uid] = v.Value
										break
									}
								}

								cell.ShowValue = finalVal
							}
						}
					} else {
						tmpDateList := strings.Split(cell.DataTime, "-")
						tmpDateValMap := make(map[string]float64)
						var newDate string
						if len(tmpDateList) == 2 {
							if dateMap, ok1 := edbMonthDataListMap[cell.EdbInfoId]; ok1 {
								if d, ok2 := dateMap[cell.DataTime]; ok2 {
									newDate = d
								}
							}
						} else {
							newDate = cell.DataTime
						}
						if !strings.Contains(cell.Value, "{") {
							cell.Value = ""
						}

						newDate, err = HandleMixTableDateChange(newDate, cell.Value)
						if err != nil {
							return
						}

						if dateValMap, ok3 := edbDayDataListMap[cell.EdbInfoId]; ok3 {
							tmpDateValMap = dateValMap
							if val, ok2 := tmpDateValMap[cell.DataTime]; ok2 {
								cellKeyVal[cell.Uid] = val
								cell.ShowValue = utils.FormatMixTableDataShowValue(val)
							}
						}
					}
				}
			case request.CustomTextDT: //自定义文本
				if cell.Value == `` {
					continue
				}
				tmpDeci, tmpErr := decimal.NewFromString(cell.Value)
				if tmpErr == nil {
					tmpVal, _ := tmpDeci.Float64()
					cellKeyVal[cell.Uid] = tmpVal
				}

			case request.FormulateCalculateDataDT: // 公式计算（A+B这种）
				calculateChainList = append(calculateChainList, cell.Uid)

			case request.InsertEdbCalculateDataDT: // 插入指标系统计算公式生成的值
				if !strings.Contains(cell.Value, "EdbInfoId") && cell.EdbInfoId > 0 {
					cell.Value, _ = fixCalculateValueConfig(cell.Value, cell.EdbInfoId)
					row[i] = cell
				}
				var cellDateTime string
				if relationConf, ok := cellRelationConfMap[cell.Uid]; ok {
					if relationConf.RelationDate.Key == `` {
						continue
					}
					relationCell, ok := cellDataRelationMap[relationConf.RelationDate.Key]
					if !ok {
						continue
					}
					cellDateTime = relationCell.DataTime
				}

				var tmpDataMap map[string]float64

				key := utils.MD5(cell.Value)
				respItemData, ok := edbSourceDataMap[key]
				if !ok {
					var tmpConfig request.CalculateConf
					err = json.Unmarshal([]byte(cell.Value), &tmpConfig)
					if err != nil {
						return
					}

					tmpDataList, ok := edbDataListMap[tmpConfig.EdbInfoId]
					if !ok {
						continue
					}
					edbInfo, ok := edbInfoMap[tmpConfig.EdbInfoId]
					if !ok {
						continue
					}

					req2 := &BaseCalculate{
						DataList:      tmpDataList,
						Frequency:     tmpConfig.Frequency,
						Formula:       tmpConfig.Formula,
						Calendar:      tmpConfig.Calendar,
						MoveType:      tmpConfig.MoveType,
						MoveFrequency: tmpConfig.MoveFrequency,
						FromFrequency: edbInfo.Frequency,
						Source:        tmpConfig.Source,
					}

					reqJson, tmpErr := json.Marshal(req2)
					if tmpErr != nil {
						utils.FileLog.Error(fmt.Sprintf("计算失败1，配置信息;%s;错误原因：%s", cell.Value, tmpErr.Error()))
						err = tmpErr
						return
					}
					respItem, tmpErr := data.BaseCalculate(string(reqJson), lang)
					if tmpErr != nil {
						utils.FileLog.Error(fmt.Sprintf("计算失败2，配置信息;%s;错误原因：%s", cell.Value, tmpErr.Error()))
						err = tmpErr
						return
					}
					if respItem.Ret != 200 {
						utils.FileLog.Error(fmt.Sprintf("计算失败3，配置信息;%s;原因：%s;错误原因：%s", cell.Value, respItem.Msg, respItem.ErrMsg))
						continue
					}

					edbSourceDataMap[key] = respItem.Data
					respItemData = respItem.Data
				}
				lenDataList := len(respItemData.DateList)
				tmpDataMap = respItemData.DataMap
				if cellDateTime == `` && lenDataList > 0 {
					cellDateTime = respItemData.DateList[lenDataList-1]
					cellDateTime, err = GetEdbDateByMoveForwardByDateList(cell.Value, respItemData.DateList)
					if err != nil {
						utils.FileLog.Error(fmt.Sprintf("日期前移失败，配置信息;%s", cell.Value))
						continue
					}
				}
				cellDateTime, err = HandleMixTableDateChange(cellDateTime, cell.Value)
				if err != nil {
					utils.FileLog.Error(fmt.Sprintf("日期变换失败，配置信息;%s， 日期：%s", cell.Value, cellDateTime))
					continue
				}
				val, ok := tmpDataMap[cellDateTime]
				if ok {
					cellKeyVal[cell.Uid] = val
					cell.ShowValue = utils.FormatMixTableDataShowValue(val)
				} else {
					cell.ShowValue = ""
				}
			case request.DateCalculateDataDT: //日期计算
				dateCalculateList = append(dateCalculateList, cell.Uid)
			}
			row[i] = cell
		}
		config[k] = row
	}

	for k, row := range config {
		for i, cell := range row {
			calculateCellMap[cell.Uid] = Cell{
				Column:   k,
				Row:      i,
				CellInfo: cell,
			}
			cell.ShowFormatValue = cell.ShowValue
			if cell.ShowStyle != `` {
				showStyleList = append(showStyleList, cell.Uid)
			}
			row[i] = cell
			cellDataRelationMap[cell.Uid] = cell
		}
		config[k] = row
	}

	if len(calculateChainList) > 0 {
		for _, cellKey := range calculateChainList {
			cellPosition, ok := calculateCellMap[cellKey]
			if !ok {
				utils.FileLog.Error("找不到单元格位置：", cellKey)
				continue
			}

			cell := config[cellPosition.Column][cellPosition.Row]
			if cell.DataType != request.FormulateCalculateDataDT { // 判断公式计算（A+B这种）类型，不是的话也过滤了
				continue
			}

			val, has, tmpErr, tmpErrMsg := getCalculateValueByCell(calculateCellMap, cellKey, cellKeyVal)
			if tmpErr != nil {
				errMsg = tmpErrMsg
				err = tmpErr
				return
			}
			if !has {
				continue
			}

			cellKeyVal[cell.Uid] = val
			cell.ShowValue = utils.FormatMixTableDataShowValue(val)
			config[cellPosition.Column][cellPosition.Row] = cell

		}
	}

	config, err, errMsg = handlerDateCalculate(dateCalculateList, calculateCellMap, config)
	if err != nil {
		return
	}

	config, err, errMsg = handleMixCellShowStyle(showStyleList, calculateCellMap, config)
	if err != nil {
		return
	}

	newMixedTableCellDataList = config

	return
}

func getCalculateValueByCell(calculateCellMap map[string]Cell, key string, cellKeyValMap map[string]float64) (val float64, has bool, err error, errMsg string) {
	val, ok := cellKeyValMap[key]
	if ok {
		has = true
		return
	}

	cell, ok := calculateCellMap[key]
	if !ok {
		err = errors.New("查找单元格" + key + "的数据失败")
		return
	}

	colData := cell.CellInfo

	if colData.DataType != request.FormulateCalculateDataDT {
		return
	}


	tagList := make([]utils.CellPosition, 0)
	var relationCellList []request.RelationCell
	if colData.Extra == `` {
		err = errors.New(colData.Uid + "没有绑定关系")
		return
	}
	err = json.Unmarshal([]byte(colData.Extra), &relationCellList)
	if err != nil {
		return
	}

	for _, relation := range relationCellList {
		tmpVal, _, tmpErr, tmpErrMsg := getCalculateValueByCell(calculateCellMap, relation.Key, cellKeyValMap)
		if tmpErr != nil {
			errMsg = tmpErrMsg
			err = tmpErr
			return
		}
		cellKeyValMap[relation.Key] = tmpVal

		rowInt, tmpErr := strconv.Atoi(relation.Row)
		if tmpErr != nil {
			err = tmpErr
			return
		}

		tagList = append(tagList, utils.CellPosition{
			Tag:   relation.Tag,
			Row:   rowInt,
			Value: tmpVal,
		})
	}

	val, errMsg, err = calculateByCellList(strings.ToUpper(colData.Value), tagList)
	if err != nil {
		return
	}
	has = true
	cellKeyValMap[key] = val

	return
}

func handleConfig(configList [][]request.MixedTableCellDataReq) (newConfig [][]request.MixedTableCellDataReq, edbInfoIdList []int, dataEdbInfoIdList []int, err error, errMsg string) {
	edbInfoIdList = make([]int, 0)
	dataEdbInfoIdList = make([]int, 0)

	for ck, rowList := range configList {
		for rk, cell := range rowList {
			switch cell.DataType {
			case request.EdbDT: // 指标信息
				edbInfoIdList = append(edbInfoIdList, cell.EdbInfoId)
			case request.InsertDataDT, request.PopInsertDataDT: // 插值、弹框插值
				dataEdbInfoIdList = append(dataEdbInfoIdList, cell.EdbInfoId)
				edbInfoIdList = append(edbInfoIdList, cell.EdbInfoId)
			case request.InsertEdbCalculateDataDT: // 插入指标计算公式生成的值
				var config request.CalculateConf
				err = json.Unmarshal([]byte(cell.Value), &config)
				if err != nil {
					return
				}
				if config.EdbInfoId == 0 && cell.EdbInfoId > 0 {
					edbInfoIdList = append(edbInfoIdList, cell.EdbInfoId)
				} else {
					edbInfoIdList = append(edbInfoIdList, config.EdbInfoId)
				}
				dataEdbInfoIdList = append(dataEdbInfoIdList, cell.EdbInfoId)

			case request.DateDT: // 日期类型
				date, newVal, tmpErr, tmpErrMsg := handleDate(cell.DataTimeType, cell.Value)
				if tmpErr != nil {
					err = tmpErr
					errMsg = tmpErrMsg
					return
				}
				rowList[rk].DataTime = date
				rowList[rk].ShowValue = date
				rowList[rk].Value = newVal //兼容原有的历史数据中系统导入日期

				if cell.DataTimeType == request.EdbDateDT {
					var config request.EdbDateConf
					err = json.Unmarshal([]byte(cell.Value), &config)
					if err != nil {
						return
					}
					edbInfoIdList = append(edbInfoIdList, config.EdbInfoId)
				}
			}
		}
		configList[ck] = rowList
	}

	newConfig = configList

	return
}

func HandleDate(dataTimeType int, val string) (date string, newVal string, err error, errMsg string) {
	return handleDate(dataTimeType, val)
}

func handleDate(dataTimeType int, val string) (date string, newVal string, err error, errMsg string) {
	newVal = val
	if val == `` {
		errMsg = "错误的日期数据"
		err = errors.New(errMsg)
		return
	}

	switch dataTimeType {
	case request.CustomDateT: //手动输入日期
		/*if !strings.Contains(val, "{") {
			newVal, err, errMsg = handleOldCustomerDateT(val)
			if err != nil {
				return
			}
		}
		date, err = HandleMixTableDateChange("", newVal)
		if err != nil {
			return
		}*/
		date = val
		return
	case request.SystemDateT: // 系统日期
		date = time.Now().Format(utils.FormatDate)
		newVal, err, errMsg = handleOldSystemDateT(val)
		if err != nil {
			return
		}
		date, err = HandleMixTableDateChange(date, newVal)
		if err != nil {
			return
		}

	case request.EdbDateDT: // 导入指标日期（指标库的最新日期）
	default:
		errMsg = "错误的日期类型"
		err = errors.New(errMsg)
		return
	}

	return
}

func handleSystemCalculateDateT(num int, frequency string) (date string, err error, errMsg string) {
	if err != nil {
		return
	}
	currDate := time.Now()
	switch frequency {
	case "", "日":
		date = currDate.AddDate(0, 0, num).Format(utils.FormatDate)
	default:
		errMsg = "错误的日期频度：" + frequency
		err = errors.New(errMsg)
		return
	}

	return
}

func handleSystemAppointDateT(currDate time.Time, appointDay, frequency string) (date string, err error, errMsg string) {
	switch frequency {
	case "本周":
		day := int(currDate.Weekday())
		if day == 0 { // 周日
			day = 7
		}
		num := 0
		switch appointDay {
		case "周一":
			num = 1
		case "周二":
			num = 2
		case "周三":
			num = 3
		case "周四":
			num = 4
		case "周五":
			num = 5
		case "周六":
			num = 6
		case "周日":
			num = 7
		}
		day = num - day
		date = currDate.AddDate(0, 0, day).Format(utils.FormatDate)
	case "本旬":
		day := currDate.Day()
		var tmpDate time.Time
		switch appointDay {
		case "第一天":
			if day <= 10 {
				tmpDate = time.Date(currDate.Year(), currDate.Month(), 1, 0, 0, 0, 0, currDate.Location())
			} else if day <= 20 {
				tmpDate = time.Date(currDate.Year(), currDate.Month(), 11, 0, 0, 0, 0, currDate.Location())
			} else {
				tmpDate = time.Date(currDate.Year(), currDate.Month(), 21, 0, 0, 0, 0, currDate.Location())
			}
		case "最后一天":
			if day <= 10 {
				tmpDate = time.Date(currDate.Year(), currDate.Month(), 10, 0, 0, 0, 0, currDate.Location())
			} else if day <= 20 {
				tmpDate = time.Date(currDate.Year(), currDate.Month(), 20, 0, 0, 0, 0, currDate.Location())
			} else {
				tmpDate = time.Date(currDate.Year(), currDate.Month()+1, 1, 0, 0, 0, 0, currDate.Location()).AddDate(0, 0, -1)
			}
		}
		date = tmpDate.Format(utils.FormatDate)
	case "本月":
		var tmpDate time.Time
		switch appointDay {
		case "第一天":
			tmpDate = time.Date(currDate.Year(), currDate.Month(), 1, 0, 0, 0, 0, currDate.Location())
		case "最后一天":
			tmpDate = time.Date(currDate.Year(), currDate.Month()+1, 1, 0, 0, 0, 0, currDate.Location()).AddDate(0, 0, -1)
		}
		date = tmpDate.Format(utils.FormatDate)
	case "本季":
		month := currDate.Month()
		var tmpDate time.Time
		switch appointDay {
		case "第一天":
			if month <= 3 {
				tmpDate = time.Date(currDate.Year(), 1, 1, 0, 0, 0, 0, currDate.Location())
			} else if month <= 6 {
				tmpDate = time.Date(currDate.Year(), 4, 1, 0, 0, 0, 0, currDate.Location())
			} else if month <= 9 {
				tmpDate = time.Date(currDate.Year(), 7, 1, 0, 0, 0, 0, currDate.Location())
			} else {
				tmpDate = time.Date(currDate.Year(), 10, 1, 0, 0, 0, 0, currDate.Location())
			}
		case "最后一天":
			if month <= 3 {
				tmpDate = time.Date(currDate.Year(), 3, 31, 0, 0, 0, 0, currDate.Location())
			} else if month <= 6 {
				tmpDate = time.Date(currDate.Year(), 6, 30, 0, 0, 0, 0, currDate.Location())
			} else if month <= 9 {
				tmpDate = time.Date(currDate.Year(), 9, 30, 0, 0, 0, 0, currDate.Location())
			} else {
				tmpDate = time.Date(currDate.Year(), 12, 31, 0, 0, 0, 0, currDate.Location())
			}
		}
		date = tmpDate.Format(utils.FormatDate)
	case "本半年":
		month := currDate.Month()
		var tmpDate time.Time
		switch appointDay {
		case "第一天":
			if month <= 6 {
				tmpDate = time.Date(currDate.Year(), 1, 1, 0, 0, 0, 0, currDate.Location())
			} else {
				tmpDate = time.Date(currDate.Year(), 7, 1, 0, 0, 0, 0, currDate.Location())
			}
		case "最后一天":
			if month <= 6 {
				tmpDate = time.Date(currDate.Year(), 6, 30, 0, 0, 0, 0, currDate.Location())
			} else {
				tmpDate = time.Date(currDate.Year(), 12, 31, 0, 0, 0, 0, currDate.Location())
			}
		}
		date = tmpDate.Format(utils.FormatDate)
	case "本年":
		var tmpDate time.Time
		switch appointDay {
		case "第一天":
			tmpDate = time.Date(currDate.Year(), 1, 1, 0, 0, 0, 0, currDate.Location())
		case "最后一天":
			tmpDate = time.Date(currDate.Year(), 12, 31, 0, 0, 0, 0, currDate.Location())
		}
		date = tmpDate.Format(utils.FormatDate)
	default:
		errMsg = "错误的日期频度：" + frequency
		err = errors.New(errMsg)
		return
	}

	return
}

func calculateByCellList(calculateFormula string, tagList []utils.CellPosition) (calVal float64, errMsg string, err error) {
	if calculateFormula == "" {
		errMsg = "公式异常"
		err = errors.New(errMsg)
		return
	}

	calculateFormula = strings.TrimPrefix(calculateFormula, "=")
	calculateFormula = strings.Replace(calculateFormula, "（", "(", -1)
	calculateFormula = strings.Replace(calculateFormula, "）", ")", -1)
	calculateFormula = strings.Replace(calculateFormula, "，", ",", -1)
	calculateFormula = strings.Replace(calculateFormula, "。", ".", -1)
	calculateFormula = strings.Replace(calculateFormula, "%", "*0.01", -1)

	rowList := make([]int, 0)
	rowListMap := make(map[int][]utils.CellPosition)
	for _, v := range tagList {
		tmpRowList, ok := rowListMap[v.Row]
		if !ok {
			rowList = append(rowList, v.Row)
			tmpRowList = make([]utils.CellPosition, 0)
		}
		tmpRowList = append(tmpRowList, v)
		rowListMap[v.Row] = tmpRowList
	}

	sort.Ints(rowList)

	list := make([]utils.CellPosition, 0)
	for _, row := range rowList {
		list = append(list, rowListMap[row]...)
	}

	formulaFormStr := utils.ReplaceFormulaByCellList(list, calculateFormula)
	if formulaFormStr == `` {
		errMsg = "公式异常"
		err = errors.New(errMsg)
		return
	}

	expression := formula.NewExpression(formulaFormStr)
	calResult, err := expression.Evaluate()
	if err != nil {
		errMsg = "计算失败"
		err = errors.New("计算失败：Err:" + err.Error() + ";formulaStr:" + formulaFormStr)
		if strings.Contains(err.Error(), "divide by zero") {
			errMsg = "分母不能为0"
			err = errors.New("分母不能为空，计算公式：" + formulaFormStr)
		}
		return
	}
	if calResult.IsNan() {
		errMsg = "计算失败"
		err = errors.New("计算失败：计算结果是:NAN;formulaStr:" + formulaFormStr)
		return
	}
	calVal, err = calResult.Float64()
	if err != nil {
		return
	}

	calVal, _ = decimal.NewFromFloat(calVal).Round(4).Float64()

	return
}

func GetEdbDateByMoveForward(conf string, edbDataList []*models.EdbDataList) (date string, err error) {
	dateList := make([]string, 0)
	for _, v := range edbDataList {
		dateList = append(dateList, v.DataTime)
	}

	date, err = GetEdbDateByMoveForwardByDateList(conf, dateList)
	return
}

func GetEdbDateByMoveForwardByDateList(conf string, dateList []string) (date string, err error) {
	moveForward := 0
	if conf != "" {
		var edbDateConf request.EdbDateChangeConf
		err = json.Unmarshal([]byte(conf), &edbDateConf)
		if err != nil {
			err = fmt.Errorf("日期变换配置json解析失败失败: %s", err.Error())
			return
		}
		moveForward = edbDateConf.MoveForward
	}
	index := len(dateList) - 1 - moveForward
	for k, v := range dateList {
		if k == index {
			date = v
			return
		}
	}
	return
}

func HandleMixTableDateChange(date, conf string) (newDate string, err error) {
	newDate = date
	if conf == "" {
		return
	}
	var edbDateConf request.EdbDateConf
	err = json.Unmarshal([]byte(conf), &edbDateConf)
	if err != nil {
		err = fmt.Errorf("日期变换配置json解析失败失败: %s, Err:%s", conf, err.Error())
		return
	}

	if newDate != "" {
		if len(edbDateConf.DateChange) > 0 {
			var dateTime time.Time
			dateTime, err = time.ParseInLocation(utils.FormatDate, newDate, time.Local)
			if err != nil {
				err = fmt.Errorf("日期解析失败: %s", err.Error())
				return
			}
			for _, v := range edbDateConf.DateChange {
				if v.ChangeType == 1 {
					dateTime = dateTime.AddDate(v.Year, v.Month, v.Day)
					newDate = dateTime.Format(utils.FormatDate)
				} else if v.ChangeType == 2 {
					newDate, err, _ = handleSystemAppointDateT(dateTime, v.FrequencyDay, v.Frequency)
					if err != nil {
						return
					}
					dateTime, err = time.ParseInLocation(utils.FormatDate, newDate, time.Local)
					if err != nil {
						err = fmt.Errorf("日期解析失败: %s", err.Error())
						return
					}
				}
			}
		}
	}

	return
}

func handleOldSystemDateT(confStr string) (newConf string, err error, errMsg string) {
	newConf = confStr
	var config request.SystemDateConf
	err = json.Unmarshal([]byte(confStr), &config)
	if err != nil {
		return
	}

	newConfig := new(request.EdbDateConf)
	dateChange := new(request.EdbDateConfDateChange)

	dateChangeList := make([]*request.EdbDateConfDateChange, 0)

	switch config.Source {
	case request.SystemCurrDateT:
		return
	case request.SystemCalculateDateT:
		dateChange.Day = config.CalculateNum
		dateChange.ChangeType = 1
		dateChangeList = append(dateChangeList, dateChange)
		newConfig.DateChange = dateChangeList
		newConfByte, e := json.Marshal(newConfig)
		if e != nil {
			err = fmt.Errorf("日期计算额外配置，json序列化失败: %s", e.Error())
			return
		}
		newConf = string(newConfByte)
		return
	case request.SystemFrequencyDateT: // 处理系统日期相关的指定频率（所在周/旬/月/季/半年/年的最后/最早一天）
		dateChange.FrequencyDay = config.Day
		dateChange.Frequency = config.Frequency
		dateChange.ChangeType = 1
		dateChangeList = append(dateChangeList, dateChange)
		newConfig.DateChange = dateChangeList
		newConfByte, e := json.Marshal(newConfig)
		if e != nil {
			err = fmt.Errorf("日期计算额外配置，json序列化失败: %s", e.Error())
			return
		}
		newConf = string(newConfByte)
		return
	default:
		return
	}

	return
}

func handlerDateCalculate(dateCalculateList []string, calculateCellMap map[string]Cell, oldConfig [][]request.MixedTableCellDataReq) (config [][]request.MixedTableCellDataReq, err error, errMsg string) {
	config = oldConfig
	if len(dateCalculateList) == 0 {
		return
	}
	if len(dateCalculateList) > 0 {
		for _, cellKey := range dateCalculateList {
			cellPosition, ok := calculateCellMap[cellKey]
			if !ok {
				utils.FileLog.Error("找不到单元格位置：", cellKey)
				continue
			}

			cell := config[cellPosition.Column][cellPosition.Row]
			if cell.DataType != request.DateCalculateDataDT { // 判断公式计算（A+B这种）类型，不是的话也过滤了
				continue
			}

			val, tmpErr, tmpErrMsg := DateCalculatePrepare(calculateCellMap, cell.Value)
			if tmpErr != nil {
				cell.ShowValue = ""
				config[cellPosition.Column][cellPosition.Row] = cell
				utils.FileLog.Error(fmt.Sprintf("%s 日期计算报错：Err：%s:%s", cellKey, tmpErr, tmpErrMsg))
				continue
			}

			cell.ShowValue = utils.FormatMixTableDataShowValue(val)
			config[cellPosition.Column][cellPosition.Row] = cell
		}
	}
	return
}

func DateCalculatePrepare(calculateCellMap map[string]Cell, config string) (val float64, err error, errMsg string) {
	var edbDateConf request.MixDateCalculateConf
	err = json.Unmarshal([]byte(config), &edbDateConf)
	if err != nil {
		err = fmt.Errorf("日期计算配置json解析失败失败: %s, Err:%s", config, err.Error())
		return
	}
	if len(edbDateConf.RelationCellList) == 0 {
		err = fmt.Errorf("日期计算 未配置日期单元格失败: %s", config)
		return
	}
	valMap := make(map[string]int)
	for _, v := range edbDateConf.RelationCellList {
		cell, ok := calculateCellMap[v.Uid]
		if !ok {
			err = fmt.Errorf("查找单元格:%s 的数据失败", v.Uid)
			return
		}
		colData := cell.CellInfo

		_, err = time.ParseInLocation(utils.FormatDate, colData.ShowValue, time.Local)
		if err != nil {
			err = fmt.Errorf("%s 的单元格非日期类型, Err: %s", colData.ShowValue, err.Error())
			return
		}
		realDiffDay := utils.GetDaysDiff1900(colData.ShowValue)

		valMap[strings.ToUpper(v.Tag)] = realDiffDay
	}

	val, errMsg, err = DateCalculateFormula(valMap, strings.ToUpper(edbDateConf.Formula))
	if err != nil {
		return
	}
	return
}

func DateCalculateFormula(valTagMap map[string]int, calculateFormula string) (calVal float64, errMsg string, err error) {
	if calculateFormula == "" {
		errMsg = "公式异常"
		err = errors.New(errMsg)
		return
	}

	calculateFormula = strings.TrimPrefix(calculateFormula, "=")
	calculateFormula = strings.Replace(calculateFormula, "（", "(", -1)
	calculateFormula = strings.Replace(calculateFormula, "）", ")", -1)
	calculateFormula = strings.Replace(calculateFormula, "，", ",", -1)
	calculateFormula = strings.Replace(calculateFormula, "。", ".", -1)
	calculateFormula = strings.Replace(calculateFormula, "%", "*0.01", -1)

	formulaFormStr := utils.ReplaceFormulaByTagMap(valTagMap, calculateFormula)

	if formulaFormStr == `` {
		errMsg = "公式异常"
		err = errors.New(errMsg)
		return
	}
	fmt.Println("公式：" + formulaFormStr)
	expression := formula.NewExpression(formulaFormStr)
	calResult, err := expression.Evaluate()
	if err != nil {
		errMsg = "计算失败"
		err = errors.New("计算失败：Err:" + err.Error() + ";formulaStr:" + formulaFormStr)
		if strings.Contains(err.Error(), "divide by zero") {
			errMsg = "分母不能为0"
			err = errors.New("分母不能为空，计算公式：" + formulaFormStr)
		}
		return
	}
	if calResult.IsNan() {
		errMsg = "计算失败"
		err = errors.New("计算失败：计算结果是:NAN;formulaStr:" + formulaFormStr)
		return
	}
	calVal, err = calResult.Float64()
	if err != nil {
		return
	}

	calVal, _ = decimal.NewFromFloat(calVal).Round(4).Float64()

	return
}

func handleMixCellShowStyle(showStyleList []string, calculateCellMap map[string]Cell, oldConfig [][]request.MixedTableCellDataReq) (config [][]request.MixedTableCellDataReq, err error, errMsg string) {
	config = oldConfig
	if len(showStyleList) == 0 {
		return
	}
	if len(showStyleList) > 0 {
		for _, cellKey := range showStyleList {
			cellPosition, ok := calculateCellMap[cellKey]
			if !ok {
				utils.FileLog.Error("找不到单元格位置：", cellKey)
				continue
			}

			cell := config[cellPosition.Column][cellPosition.Row]
			val := cell.ShowValue
			if cell.ShowValue == "" {
				_, err := strconv.ParseFloat(cell.Value, 64)
				if err == nil {
					val = cell.Value
					cell.ShowValue = cell.Value
				}
			}
			_, e := strconv.ParseFloat(val, 64) // 将字符串转换成float类型
			if e != nil {                       // 如果没有错误发生则返回true，说明该字符串是一个合法的数字
				continue
			}
			var styleConf request.MixCellShowStyle
			err = json.Unmarshal([]byte(cell.ShowStyle), &styleConf)
			if err != nil {
				err = fmt.Errorf("日期计算配置json解析失败失败: %s, Err:%s", config, err.Error())
				return
			}
			hasPercent := false
			if styleConf.Nt == "percent" {
				hasPercent = true
			}

			if styleConf.Pn != 0 || styleConf.Nt != "" {
				if styleConf.Pn != 0 {
					if styleConf.Decimal == nil {
						styleConf.Decimal = new(int)
					}
					*styleConf.Decimal += styleConf.Pn
					if *styleConf.Decimal < 0 {
						*styleConf.Decimal = 0
					}
					cell.ShowFormatValue = utils.RoundNumber(cell.ShowValue, *styleConf.Decimal, hasPercent)
					styleConf.Pn = 0
				} else if styleConf.Decimal != nil {
					cell.ShowFormatValue = utils.RoundNumber(cell.ShowValue, *styleConf.Decimal, hasPercent)
				} else {
					if hasPercent {
						numDecimal, _ := decimal.NewFromString(cell.ShowValue)
						tmpStr := numDecimal.Mul(decimal.NewFromInt(100)).String()
						cell.ShowFormatValue = tmpStr + "%"
					} else {
						cell.ShowFormatValue = cell.ShowValue
					}
				}
			} else if styleConf.Decimal != nil {
				cell.ShowFormatValue = utils.RoundNumber(cell.ShowValue, *styleConf.Decimal, hasPercent)
			} else {
				cell.ShowFormatValue = cell.ShowValue
			}
			styleConf.GlObj = nil
			tmpStyleConf, err := json.Marshal(styleConf)
			if err == nil {
				cell.ShowStyle = string(tmpStyleConf)
			}
			config[cellPosition.Column][cellPosition.Row] = cell
		}
	}
	return
}

func changePointDecimalPlaces(str string, changeNum int, numberType string, isPercent bool) (newStr string) {
	newStr = str
	var decimalPlaces int
	dotIndex := strings.Index(newStr, ".") // 查找小数点的位置
	if dotIndex == -1 {
		decimalPlaces = 0
	} else {
		decimalPlaces = len(newStr) - dotIndex - 1
	}
	val, _ := strconv.ParseFloat(str, 64)
	if isPercent {
		if numberType == "number" { //百分数转成小数
			val = val / 100
			if decimalPlaces > 2 {
				decimalPlaces += 2
			} else if decimalPlaces == 1 {
				decimalPlaces += 1
			} else if decimalPlaces == 0 {
				if len(str) == 1 {
					decimalPlaces = 2
				} else if len(str) == 2 {
					if str[1] == '0' {
						decimalPlaces = 1
					} else {
						decimalPlaces = 2
					}
				}
			}
			isPercent = false
		}
	} else {
		if numberType == "percent" {
			if decimalPlaces > 2 {
				decimalPlaces -= 2
			} else if decimalPlaces == 1 {
				decimalPlaces = 0
			}
			val = val * 100
		}
	}
	if decimalPlaces > 0 {
		val, _ = decimal.NewFromFloat(val).Round(int32(decimalPlaces)).Float64()
		newStr = strconv.FormatFloat(val, 'f', decimalPlaces, 64)
	} else {
		newStr = fmt.Sprintf("%.f", val)
	}
	decimalPlaces = 0
	dotIndex = strings.Index(newStr, ".") // 查找小数点的位置
	if dotIndex == -1 {
		decimalPlaces = 0
	} else {
		decimalPlaces = len(newStr) - dotIndex - 1
	}
	decimalPlaces += changeNum

	if decimalPlaces < 0 {
		decimalPlaces = 0
	}
	val, _ = decimal.NewFromFloat(val).Round(int32(decimalPlaces)).Float64()
	newStr = strconv.FormatFloat(val, 'f', decimalPlaces, 64)
	if numberType == "percent" || isPercent {
		newStr += "%"
	}
	return
}

func fixCalculateValueConfig(conf string, edbInfoId int) (newConf string, err error) {
	newConf = conf
	if edbInfoId == 0 {
		return
	}
	var tmpConfig request.CalculateConf
	err = json.Unmarshal([]byte(conf), &tmpConfig)
	if err != nil {
		return
	}
	if tmpConfig.EdbInfoId == 0 {
		tmpConfig.EdbInfoId = edbInfoId
		newConfByte, _ := json.Marshal(tmpConfig)
		newConf = string(newConfByte)
		return
	}
	return
}
